/*-
# X-BASED MISSING LINK(tm)
#
#  Mlink.c
#
###
#
#  Copyright (c) 1994 - 2003	David Albert Bagley, bagleyd@tux.org
#
#                   All Rights Reserved
#
#  Permission to use, copy, modify, and distribute this software and
#  its documentation for any purpose and without fee is hereby granted,
#  provided that the above copyright notice appear in all copies and
#  that both that copyright notice and this permission notice appear in
#  supporting documentation, and that the name of the author not be
#  used in advertising or publicity pertaining to distribution of the
#  software without specific, written prior permission.
#
#  This program is distributed in the hope that it will be "playable",
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
*/

/* Methods file for Mlink */

#include <stdio.h>
#include <stdlib.h>
#ifdef VMS
#include <unixlib.h>
#else
#if HAVE_UNISTD_H
#include <unistd.h>
#endif
#endif
#include <X11/IntrinsicP.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/CoreP.h>
#include "MlinkP.h"

#ifndef LOGPATH
#ifdef VMS
#define LOGPATH "SYS$SCRATCH:"
#else
#define LOGPATH "/usr/tmp"
#endif
#endif
#ifndef LOGFILE
#define LOGFILE "mlink.log"
#endif

static void InitializeMlink(Widget request, Widget renew);
static void ExposeMlink(Widget renew, XEvent * event, Region region);
static void ResizeMlink(MlinkWidget w);
static void DestroyMlink(Widget old);
static Boolean SetValuesMlink(Widget current, Widget request, Widget renew);
static void QuitMlink(MlinkWidget w, XEvent * event, char **args, int nArgs);
static void HideMlink(MlinkWidget w, XEvent * event, char **args, int nArgs);
static void MoveMlinkTop(MlinkWidget w, XEvent * event, char **args, int nArgs);
static void MoveMlinkLeft(MlinkWidget w, XEvent * event, char **args, int nArgs);
static void MoveMlinkRight(MlinkWidget w, XEvent * event, char **args, int nArgs);
static void MoveMlinkBottom(MlinkWidget w, XEvent * event, char **args, int nArgs);
static void MoveMlinkInput(MlinkWidget w, int x, int direction, int shift, int control);
static void SelectMlink(MlinkWidget w, XEvent * event, char **args, int nArgs);
static void ReleaseMlink(MlinkWidget w, XEvent * event, char **args, int nArgs);
static void RandomizeMlink(MlinkWidget w, XEvent * event, char **args, int nArgs);
static void RandomizeMlinkMaybe(MlinkWidget w, XEvent * event, char **args, int nArgs);
static void RandomizeMlink2(MlinkWidget w, XEvent * event, char **args, int nArgs);
static void GetMlink(MlinkWidget w, XEvent * event, char **args, int nArgs);
static void WriteMlink(MlinkWidget w, XEvent * event, char **args, int nArgs);
static void ClearMlink(MlinkWidget w, XEvent * event, char **args, int nArgs);
static void UndoMlink(MlinkWidget w, XEvent * event, char **args, int nArgs);
static void SolveMlink(MlinkWidget w, XEvent * event, char **args, int nArgs);
static void OrientizeMlink(MlinkWidget w, XEvent * event, char **args, int nArgs);
static void MiddleMlink(MlinkWidget w, XEvent * event, char **args, int nArgs);
static void EnterMlink(MlinkWidget w, XEvent * event, char **args, int nArgs);
static void LeaveMlink(MlinkWidget w, XEvent * event, char **args, int nArgs);
static void ControlMlink(MlinkWidget w, int direction);
static void RotateMlink(MlinkWidget w, int direction, int tile);
static void SlideMlink(MlinkWidget w, int direction);

static int  SelectTiles(MlinkWidget w, int x, int y, int *i, int *j);
static void SelectSlideTiles(MlinkWidget w, int pos);
static void SetAllColors(MlinkWidget w, Boolean init);
static void GetColor(MlinkWidget w, int face, Boolean init);
static void MoveShiftCb(MlinkWidget w, int direction, int fast);
static void CheckTiles(MlinkWidget w);
static void ResetTiles(MlinkWidget w);
static void ResizeTiles(MlinkWidget w);
static void MoveNoTiles(MlinkWidget w);
static Boolean MoveTilesDir(MlinkWidget w, int direction, int tile, Boolean all, int fast);
static void GetTiles(MlinkWidget w);
static void WriteTiles(MlinkWidget w);
static void ClearTiles(MlinkWidget w);
static void UndoTiles(MlinkWidget w);
static void RandomizeTiles(MlinkWidget w);
static void SolveTiles(MlinkWidget w);
static void OrientizeTiles(MlinkWidget w);
static void MiddleTiles(MlinkWidget w);
static void MoveTiles(MlinkWidget w, int from, int fast);
static void RotateTiles(MlinkWidget w, int start, int finish, int direction, int fast);
static int  ExchangeTiles(MlinkWidget w, int pos1, int pos2);
static int  TileNFromSpace(MlinkWidget w, int n, int direction);
static void DrawFrame(MlinkWidget w, GC gc);
static void DrawAllTilesForColumns(MlinkWidget w, int start, int finish);
static void DrawTile(MlinkWidget w, int pos, Boolean blank, Boolean erase,
	int pressedOffset, int offsetX, int offsetY);
static void DrawLink(MlinkWidget w, GC gc, int pos, int offsetX, int offsetY);
static Pixel brighter(MlinkWidget w, Pixel pixel);
static Pixel darker(MlinkWidget w, Pixel pixel);
static void fill3DRect(MlinkWidget w, GC gc,
 GC darkerGC, GC brighterGC,
 int x, int y, int width, int height, Boolean raised);
static void DiscreteMoves(MlinkWidget w, int from, int dir);
static Boolean PositionTile(MlinkWidget w, int x, int *r);
static int  Column(MlinkWidget w, int pos);
static int  Row(MlinkWidget w, int pos);
static int  int2String(char *buf, int number, int base, Boolean capital);

static char defaultTranslationsMlink[] =
"<KeyPress>q: Quit()\n\
 Ctrl<KeyPress>C: Quit()\n\
 <KeyPress>osfCancel: Hide()\n\
 <KeyPress>Escape: Hide()\n\
 <KeyPress>osfEscape: Hide()\n\
 Ctrl<KeyPress>[: Hide()\n\
 <KeyPress>0x1B: Hide()\n\
 <KeyPress>Up: MoveTop()\n\
 <KeyPress>osfUp: MoveTop()\n\
 <KeyPress>KP_Up: MoveTop()\n\
 <KeyPress>KP_8: MoveTop()\n\
 <KeyPress>R8: MoveTop()\n\
 <KeyPress>Left: MoveLeft()\n\
 <KeyPress>osfLeft: MoveLeft()\n\
 <KeyPress>KP_Left: MoveLeft()\n\
 <KeyPress>KP_4: MoveLeft()\n\
 <KeyPress>R10: MoveLeft()\n\
 <KeyPress>Right: MoveRight()\n\
 <KeyPress>osfRight: MoveRight()\n\
 <KeyPress>KP_Right: MoveRight()\n\
 <KeyPress>KP_6: MoveRight()\n\
 <KeyPress>R12: MoveRight()\n\
 <KeyPress>Down: MoveBottom()\n\
 <KeyPress>osfDown: MoveBottom()\n\
 <KeyPress>KP_Down: MoveBottom()\n\
 <KeyPress>KP_2: MoveBottom()\n\
 <KeyPress>R14: MoveBottom()\n\
 <Btn1Down>: Select()\n\
 <Btn1Up>: Release()\n\
 <KeyPress>r: Randomize()\n\
 <Btn3Down>: RandomizeMaybe()\n\
 <Btn3Down>(2+): Randomize2()\n\
 <KeyPress>g: Get()\n\
 <KeyPress>w: Write()\n\
 <KeyPress>c: Clear()\n\
 <KeyPress>u: Undo()\n\
 <KeyPress>s: Solve()\n\
 <KeyPress>o: Orientize()\n\
 <KeyPress>m: Middle()\n\
 <EnterWindow>: Enter()\n\
 <LeaveWindow>: Leave()";

static XtActionsRec actionsListMlink[] =
{
	{(char *) "Quit", (XtActionProc) QuitMlink},
	{(char *) "Hide", (XtActionProc) HideMlink},
	{(char *) "MoveTop", (XtActionProc) MoveMlinkTop},
	{(char *) "MoveLeft", (XtActionProc) MoveMlinkLeft},
	{(char *) "MoveRight", (XtActionProc) MoveMlinkRight},
	{(char *) "MoveBottom", (XtActionProc) MoveMlinkBottom},
	{(char *) "Select", (XtActionProc) SelectMlink},
	{(char *) "Release", (XtActionProc) ReleaseMlink},
	{(char *) "Randomize", (XtActionProc) RandomizeMlink},
	{(char *) "RandomizeMaybe", (XtActionProc) RandomizeMlinkMaybe},
	{(char *) "Randomize2", (XtActionProc) RandomizeMlink2},
	{(char *) "Get", (XtActionProc) GetMlink},
	{(char *) "Write", (XtActionProc) WriteMlink},
	{(char *) "Clear", (XtActionProc) ClearMlink},
	{(char *) "Undo", (XtActionProc) UndoMlink},
	{(char *) "Solve", (XtActionProc) SolveMlink},
	{(char *) "Orientize", (XtActionProc) OrientizeMlink},
	{(char *) "Middle", (XtActionProc) MiddleMlink},
	{(char *) "Enter", (XtActionProc) EnterMlink},
	{(char *) "Leave", (XtActionProc) LeaveMlink}
};

static XtResource resourcesMlink[] =
{
	{XtNuserName, XtCUserName, XtRString, sizeof (String),
	 XtOffset(MlinkWidget, mlink.username),
	 XtRString, (caddr_t) "nobody"},
	{XtNfaceColor0, XtCLabel, XtRString, sizeof (String),
	 XtOffset(MlinkWidget, mlink.faceName[0]),
	 XtRString, (caddr_t) "White"},
	{XtNfaceColor1, XtCLabel, XtRString, sizeof (String),
	 XtOffset(MlinkWidget, mlink.faceName[1]),
	 XtRString, (caddr_t) "Yellow"},
	{XtNfaceColor2, XtCLabel, XtRString, sizeof (String),
	 XtOffset(MlinkWidget, mlink.faceName[2]),
	 XtRString, (caddr_t) "Green"},
	{XtNfaceColor3, XtCLabel, XtRString, sizeof (String),
	 XtOffset(MlinkWidget, mlink.faceName[3]),
	 XtRString, (caddr_t) "Red"},
	{XtNfaceColor4, XtCLabel, XtRString, sizeof (String),
	 XtOffset(MlinkWidget, mlink.faceName[4]),
	 XtRString, (caddr_t) "Blue"},
	{XtNfaceColor5, XtCLabel, XtRString, sizeof (String),
	 XtOffset(MlinkWidget, mlink.faceName[5]),
	 XtRString, (caddr_t) "Magenta"},
	{XtNfaceColor6, XtCLabel, XtRString, sizeof (String),
	 XtOffset(MlinkWidget, mlink.faceName[6]),
	 XtRString, (caddr_t) "Cyan"},
	{XtNfaceColor7, XtCLabel, XtRString, sizeof (String),
	 XtOffset(MlinkWidget, mlink.faceName[7]),
	 XtRString, (caddr_t) "Orange"},
	{XtNforeground, XtCForeground, XtRPixel, sizeof (Pixel),
	 XtOffset(MlinkWidget, mlink.foreground),
	 XtRString, (caddr_t) "Cyan" /*XtDefaultForeground*/},
	{XtNbackground, XtCBackground, XtRPixel, sizeof (Pixel),
	 XtOffset(MlinkWidget, mlink.background),
	 XtRString, (caddr_t) "NavyBlue" /*XtDefaultBackground*/},
	{XtNtileColor, XtCBackground, XtRPixel, sizeof (Pixel),
	 XtOffset(MlinkWidget, mlink.tileColor),
	 XtRString, (caddr_t) "Black" /*XtDefaultForeground*/},
	{XtNtileBorder, XtCForeground, XtRPixel, sizeof (Pixel),
	 XtOffset(MlinkWidget, mlink.borderColor),
	 XtRString, (caddr_t) "gray25" /*XtDefaultBackground*/},
	{XtNwidth, XtCWidth, XtRDimension, sizeof (Dimension),
	 XtOffset(MlinkWidget, core.width),
	 XtRString, (caddr_t) "200"},
	{XtNheight, XtCHeight, XtRDimension, sizeof (Dimension),
	 XtOffset(MlinkWidget, core.height),
	 XtRString, (caddr_t) "200"},
	{XtNtiles, XtCTiles, XtRInt, sizeof (int),
	 XtOffset(MlinkWidget, mlink.tiles),
	 XtRString, (caddr_t) "4"},	/* DEFAULTTILES */
	{XtNfaces, XtCFaces, XtRInt, sizeof (int),
	 XtOffset(MlinkWidget, mlink.faces),
	 XtRString, (caddr_t) "4"},	/* DEFAULTFACES */
	{XtNorient, XtCOrient, XtRBoolean, sizeof (Boolean),
	 XtOffset(MlinkWidget, mlink.orient),
	 XtRString, (caddr_t) "FALSE"},	/*DEFAULTORIENT */
	{XtNmiddle, XtCMiddle, XtRBoolean, sizeof (Boolean),
	 XtOffset(MlinkWidget, mlink.middle),
	 XtRString, (caddr_t) "TRUE"},	/*DEFAULTMIDDLE */
	{XtNmono, XtCMono, XtRBoolean, sizeof (Boolean),
	 XtOffset(MlinkWidget, mlink.mono),
	 XtRString, (caddr_t) "FALSE"},
	{XtNreverse, XtCReverse, XtRBoolean, sizeof (Boolean),
	 XtOffset(MlinkWidget, mlink.reverse),
	 XtRString, (caddr_t) "FALSE"},
	{XtNmoveDelay, XtCMoveDelay, XtRInt, sizeof (int),
	 XtOffset(MlinkWidget, mlink.moveDelay),
	 XtRString, (caddr_t) "10"},
	{XtNbase, XtCBase, XtRInt, sizeof (int),
	 XtOffset(MlinkWidget, mlink.base),
	 XtRString, (caddr_t) "10"},
	{XtNstart, XtCBoolean, XtRBoolean, sizeof (Boolean),
	 XtOffset(MlinkWidget, mlink.started),
	 XtRString, (caddr_t) "FALSE"},
	{XtNcheat, XtCBoolean, XtRBoolean, sizeof (Boolean),
	 XtOffset(MlinkWidget, mlink.cheat),
	 XtRString, (caddr_t) "FALSE"},
	{XtNmenu, XtCMenu, XtRInt, sizeof (int),
	 XtOffset(MlinkWidget, mlink.menu),
	 XtRString, (caddr_t) "-1"},
	{XtNselectCallback, XtCCallback, XtRCallback, sizeof (caddr_t),
	 XtOffset(MlinkWidget, mlink.select),
	 XtRCallback, (caddr_t) NULL}
};

MlinkClassRec mlinkClassRec =
{
	{
		(WidgetClass) & widgetClassRec,		/* superclass */
		(char *) "Mlink",	/* class name */
		sizeof (MlinkRec),	/* widget size */
		NULL,		/* class initialize */
		NULL,		/* class part initialize */
		FALSE,		/* class inited */
		(XtInitProc) InitializeMlink,	/* initialize */
		NULL,		/* initialize hook */
		XtInheritRealize,	/* realize */
		actionsListMlink,	/* actions */
		XtNumber(actionsListMlink),	/* num actions */
		resourcesMlink,	/* resources */
		XtNumber(resourcesMlink),	/* num resources */
		NULLQUARK,	/* xrm class */
		TRUE,		/* compress motion */
		TRUE,		/* compress exposure */
		TRUE,		/* compress enterleave */
		TRUE,		/* visible interest */
		(XtWidgetProc) DestroyMlink,	/* destroy */
		(XtWidgetProc) ResizeMlink,	/* resize */
		(XtExposeProc) ExposeMlink,	/* expose */
		(XtSetValuesFunc) SetValuesMlink,	/* set values */
		NULL,		/* set values hook */
		XtInheritSetValuesAlmost,	/* set values almost */
		NULL,		/* get values hook */
		NULL,		/* accept focus */
		XtVersion,	/* version */
		NULL,		/* callback private */
		defaultTranslationsMlink,	/* tm table */
		NULL,		/* query geometry */
		NULL,		/* display accelerator */
		NULL		/* extension */
	},
	{
		0		/* ignore */
	}
};

WidgetClass mlinkWidgetClass = (WidgetClass) & mlinkClassRec;

#ifndef HAVE_USLEEP
#if !defined( VMS ) || defined( XVMSUTILS ) ||  ( __VMS_VER >= 70000000 )
#ifdef USE_XVMSUTILS
#include <X11/unix_time.h>
#endif
#if HAVE_SYS_TIME_H
#include <sys/time.h>
#else
#if HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif
#endif
#endif
#if defined(SYSV) || defined(SVR4)
#ifdef LESS_THAN_AIX3_2
#include <sys/poll.h>
#else /* !LESS_THAN_AIX3_2 */
#include <poll.h>
#endif /* !LESS_THAN_AIX3_2 */
#endif /* defined(SYSV) || defined(SVR4) */

/* not static in case usleep found in system include */
int
usleep(unsigned int usec)
{
#if (defined (SYSV) || defined(SVR4)) && !defined(__hpux)
#if defined(HAVE_NANOSLEEP)
	{
		struct timespec rqt;

		rqt.tv_nsec = 1000 * (usec % (unsigned int) 1000000);
		rqt.tv_sec = usec / (unsigned int) 1000000;
		return nanosleep(&rqt, NULL);
	}
#else
	(void) poll(
#if defined(__cplusplus) || defined(c_plusplus)
		(pollfd *) /* guess */
#else
		(void *)
#endif
		0, (int) 0, usec / 1000);	/* ms resolution */
#endif
#else
#ifdef VMS
	long        timadr[2];

	if (usec != 0) {
		timadr[0] = -usec * 10;
		timadr[1] = -1;

		sys$setimr(4, &timadr, 0, 0, 0);
		sys$waitfr(4);
	}
#else
	struct timeval time_out;

#if  0
	/* (!defined(AIXV3) && !defined(__hpux)) */
	extern int  select(int, fd_set *, fd_set *, fd_set *, struct timeval *);

#endif

	time_out.tv_usec = usec % (unsigned int) 1000000;
	time_out.tv_sec = usec / (unsigned int) 1000000;
	(void) select(0, (void *) 0, (void *) 0, (void *) 0, &time_out);
#endif
#endif
	return 0;
}
#endif

static void
Sleep(unsigned int cMilliseconds)
{
	(void) usleep(cMilliseconds * 1000);
}

void
intCat(char ** string, const char * var1, const int var2)
{
	if (!(*string = (char *) malloc(strlen(var1) + 21)))
        	XtError("Not enough memory, exiting.");
	(void) sprintf(*string, "%s%d", var1, var2);
}

void
stringCat(char ** string, const char * var1, const char * var2)
{
	if (!(*string = (char *) malloc(strlen(var1) + strlen(var2) + 1)))
        	XtError("Not enough memory, exiting.");
	(void) sprintf(*string, "%s%s", var1, var2);
}

static void
InitializeMlink(Widget request, Widget renew)
{
	MlinkWidget w = (MlinkWidget) renew;

	w->mlink.tileOfPosition = NULL;
	CheckTiles(w);
	InitMoves();
	ResetTiles(w);
	w->mlink.numSlices = ((w->mlink.moveDelay < MAXSLICES) ?
		w->mlink.moveDelay + 1 : MAXSLICES);
	(void) SRAND(getpid());
	w->mlink.cheat = False;
	w->mlink.depth = DefaultDepthOfScreen(XtScreen(w));
	SetAllColors(w, True);
	ResizeMlink(w);
}

static void
DestroyMlink(Widget old)
{
	MlinkWidget w = (MlinkWidget) old;
	int         face;

	for (face = 0; face < MAXFACES; face++)
		XtReleaseGC(old, w->mlink.faceGC[face]);
	XtReleaseGC(old, w->mlink.tileGC);
	XtReleaseGC(old, w->mlink.tileBrighterGC);
	XtReleaseGC(old, w->mlink.tileDarkerGC);
	XtReleaseGC(old, w->mlink.puzzleGC);
	XtReleaseGC(old, w->mlink.inverseGC);
	XtRemoveCallbacks(old, XtNselectCallback, w->mlink.select);
}

static void
ResizeMlink(MlinkWidget w)
{
	w->mlink.delta.x = 3;
	w->mlink.delta.y = 3;
	w->mlink.offset.x = MAX(((int) w->core.width - w->mlink.delta.x) /
				w->mlink.tiles, 0);
	w->mlink.offset.y = MAX(((int) w->core.height - w->mlink.delta.y) /
				w->mlink.faces, 0);
	w->mlink.faceSize.x = w->mlink.offset.x * w->mlink.tiles +
		w->mlink.delta.x;
	w->mlink.faceSize.y = w->mlink.offset.y + w->mlink.delta.y;
	w->mlink.puzzleSize.x = w->mlink.faceSize.x + w->mlink.delta.x;
	w->mlink.puzzleSize.y = (w->mlink.faceSize.y - w->mlink.delta.y) *
		w->mlink.faces + w->mlink.delta.y;
	w->mlink.puzzleOffset.x = ((int) w->core.width -
				   w->mlink.puzzleSize.x + 2) / 2;
	w->mlink.puzzleOffset.y = ((int) w->core.height -
				   w->mlink.puzzleSize.y + 2) / 2;
	w->mlink.tileSize.x = MAX(w->mlink.offset.x - w->mlink.delta.x, 0);
	w->mlink.tileSize.y = MAX(w->mlink.offset.y - w->mlink.delta.y, 0);
	ResizeTiles(w);
}

static void
ExposeMlink(Widget renew, XEvent * event, Region region)
{
	MlinkWidget w = (MlinkWidget) renew;

	if (w->core.visible) {
		XFillRectangle(XtDisplay(w), XtWindow(w), w->mlink.inverseGC,
			0, 0, w->core.width, w->core.height);
		DrawFrame(w, (w->mlink.focus) ? w->mlink.puzzleGC :
			w->mlink.inverseGC);
		DrawAllTiles(w);
	}
}

static      Boolean
SetValuesMlink(Widget current, Widget request, Widget renew)
{
	MlinkWidget c = (MlinkWidget) current, w = (MlinkWidget) renew;
	Boolean     redraw = False, setColors = False;
	Boolean     redrawTiles = False;
	int         face;

	CheckTiles(w);
	for (face = 0; face < MAXFACES; face++) {
		if (strcmp(w->mlink.faceName[face], c->mlink.faceName[face])) {
			setColors = True;
			break;
		}
	}
	if (w->mlink.background != c->mlink.background ||
	    w->mlink.foreground != c->mlink.foreground ||
	    w->mlink.borderColor != c->mlink.borderColor ||
	    w->mlink.tileColor != c->mlink.tileColor ||
	    w->mlink.reverse != c->mlink.reverse ||
	    w->mlink.mono != c->mlink.mono ||
	    setColors) {
		SetAllColors(w, False);
		redrawTiles = True;
	}
	if (w->mlink.tiles != c->mlink.tiles ||
	    w->mlink.faces != c->mlink.faces ||
	    w->mlink.orient != c->mlink.orient ||
	    w->mlink.middle != c->mlink.middle ||
	    w->mlink.base != c->mlink.base) {
		ResetTiles(w);
		ResizeMlink(w);
		redraw = True;
	} else if (w->mlink.offset.x != c->mlink.offset.x ||
		   w->mlink.offset.y != c->mlink.offset.y) {
		ResizeMlink(w);
		redraw = True;
	}
	if (w->mlink.moveDelay != c->mlink.moveDelay) {
		w->mlink.numSlices = ((w->mlink.moveDelay < MAXSLICES) ?
			w->mlink.moveDelay + 1 : MAXSLICES);
	}
	if (w->mlink.menu != -1) {
		switch (w->mlink.menu) {
		case 0:
			w->mlink.menu = -1;
			GetTiles(w);
			break;
		case 1:
			w->mlink.menu = -1;
			WriteTiles(w);
			break;
		case 3:
			w->mlink.menu = -1;
			ClearTiles(w);
			break;
		case 4:
			w->mlink.menu = -1;
			UndoTiles(w);
			break;
		case 5:
			w->mlink.menu = -1;
			RandomizeTiles(w);
			break;
		case 6:
			w->mlink.menu = -1;
			SolveTiles(w);
			break;
		case 7:
			w->mlink.menu = -1;
			OrientizeTiles(w);
			break;
		case 8:
			w->mlink.menu = -1;
			MiddleTiles(w);
			break;
		default:
			w->mlink.menu = -1;
			break;
		}
	}
	if (redrawTiles && !redraw && XtIsRealized(renew) && renew->core.visible) {
		DrawFrame(w, c->mlink.inverseGC);
		if (w->mlink.focus)
			DrawFrame(w, w->mlink.puzzleGC);
		DrawAllTiles(w);
	}
	return (redraw);
}

static void
QuitMlink(MlinkWidget w, XEvent * event, char **args, int nArgs)
{
	XtCloseDisplay(XtDisplay(w));
	exit(0);
}

static void
HideMlink(MlinkWidget w, XEvent * event, char **args, int nArgs)
{
	mlinkCallbackStruct cb;

	cb.reason = MLINK_HIDE;
	XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
}

static void
SelectMlink(MlinkWidget w, XEvent * event, char **args, int nArgs)
{
	int         i, j, pos, shift;

	pos = SelectTiles(w, event->xbutton.x, event->xbutton.y, &i, &j);
	if (-w->mlink.tileFaces != pos) {
		w->mlink.currentTile = i;
		w->mlink.currentFace = j;
		w->mlink.currentRef = pos;
		shift = (int) (event->xkey.state & (ShiftMask | LockMask));
		if (shift || !CheckSolved(w)) {
			pos = w->mlink.currentTile + w->mlink.currentFace * w->mlink.tiles;
			DrawTile(w, pos, (w->mlink.tileOfPosition[i + j * w->mlink.tiles] <= 0),
				 False, TRUE, 0, 0);
		}
	} else
		w->mlink.currentRef = -w->mlink.tileFaces;
}

static void
ReleaseMlink(MlinkWidget w, XEvent * event, char **args, int nArgs)
{
	int         i, j, pos, diff, shift;

	if (w->mlink.currentRef == -w->mlink.tileFaces)
		return;
	pos = w->mlink.currentTile + w->mlink.currentFace * w->mlink.tiles;
	DrawTile(w, pos, (w->mlink.tileOfPosition[pos] <= 0), True, TRUE, 0, 0);
	if (w->mlink.tileOfPosition[pos] > 0) {
		DrawTile(w, pos, False, False, FALSE, 0, 0);
	}
	shift = (int) (event->xkey.state & (ShiftMask | LockMask));
	if (!shift && CheckSolved(w))
		MoveNoTiles(w);
	else {
		pos = SelectTiles(w, event->xbutton.x, event->xbutton.y, &i, &j);
		if (-w->mlink.tileFaces != pos) {
			if (j == w->mlink.currentFace) {
				pos = w->mlink.currentRef;
				if (pos / w->mlink.tiles == 0 &&
				    j == Row(w, w->mlink.spacePosition) && pos != 0) {
					if (shift && CheckSolved(w))
						MoveNoTiles(w);
					else {
						SelectSlideTiles(w, pos);
						w->mlink.currentTile = w->mlink.tiles;
						if (CheckSolved(w)) {
							mlinkCallbackStruct cb;

							cb.reason = MLINK_SOLVED;
							XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
						}
					}
				}
			} else {
				diff = (w->mlink.currentFace - j + w->mlink.faces) % w->mlink.faces;
				if (diff > w->mlink.faces / 2)
					for (i = 0; i < w->mlink.faces - diff; i++)
						(void) MoveMlink(w, BOTTOM, w->mlink.currentTile, shift, NORMAL);
				else
					for (i = 0; i < diff; i++)
						(void) MoveMlink(w, TOP, w->mlink.currentTile, shift, NORMAL);
				if (!shift && CheckSolved(w)) {
					mlinkCallbackStruct cb;

					cb.reason = MLINK_SOLVED;
					XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
				}
			}
		}
	}
	w->mlink.currentRef = -w->mlink.tileFaces;
}

static void
RandomizeMlink(MlinkWidget w, XEvent * event, char **args, int nArgs)
{
	RandomizeTiles(w);
}

static void
RandomizeMlinkMaybe(MlinkWidget w, XEvent * event, char **args, int nArgs)
{
	if (!w->mlink.started)
		RandomizeTiles(w);
#ifdef HAVE_MOTIF
	else {
		mlinkCallbackStruct cb;

		cb.reason = MLINK_RANDOMIZE_QUERY;
		XtCallCallbacks((Widget) w, (char *) XtNselectCallback, &cb);
	}
#endif
}

static void
RandomizeMlink2(MlinkWidget w, XEvent * event, char **args, int nArgs)
{
#ifdef HAVE_MOTIF
	if (!w->mlink.started)
#endif
		RandomizeTiles(w);
}

static void
GetMlink(MlinkWidget w, XEvent * event, char **args, int nArgs)
{
	GetTiles(w);
}

static void
WriteMlink(MlinkWidget w, XEvent * event, char **args, int nArgs)
{
	WriteTiles(w);
}

static void
ClearMlink(MlinkWidget w, XEvent * event, char **args, int nArgs)
{
	ClearTiles(w);
}

static void
UndoMlink(MlinkWidget w, XEvent * event, char **args, int nArgs)
{
	UndoTiles(w);
}

static void
SolveMlink(MlinkWidget w, XEvent * event, char **args, int nArgs)
{
	SolveTiles(w);
}

static void
OrientizeMlink(MlinkWidget w, XEvent * event, char **args, int nArgs)
{
	OrientizeTiles(w);
}

static void
MiddleMlink(MlinkWidget w, XEvent * event, char **args, int nArgs)
{
	MiddleTiles(w);
}

static void
EnterMlink(MlinkWidget w, XEvent * event, char **args, int nArgs)
{
	w->mlink.focus = True;
	DrawFrame(w, w->mlink.puzzleGC);
}

static void
LeaveMlink(MlinkWidget w, XEvent * event, char **args, int nArgs)
{
	w->mlink.focus = False;
	DrawFrame(w, w->mlink.inverseGC);
}

static void
MoveMlinkTop(MlinkWidget w, XEvent * event, char **args, int nArgs)
{
	MoveMlinkInput(w, event->xbutton.x, TOP,
		       (int) (event->xkey.state & (ShiftMask | LockMask)),
		       (int) (event->xkey.state & ControlMask));
}

static void
MoveMlinkLeft(MlinkWidget w, XEvent * event, char **args, int nArgs)
{
	MoveMlinkInput(w, event->xbutton.x, LEFT,
		       (int) (event->xkey.state & (ShiftMask | LockMask)),
		       (int) (event->xkey.state & ControlMask));
}

static void
MoveMlinkRight(MlinkWidget w, XEvent * event, char **args, int nArgs)
{
	MoveMlinkInput(w, event->xbutton.x, RIGHT,
		       (int) (event->xkey.state & (ShiftMask | LockMask)),
		       (int) (event->xkey.state & ControlMask));
}

static void
MoveMlinkBottom(MlinkWidget w, XEvent * event, char **args, int nArgs)
{
	MoveMlinkInput(w, event->xbutton.x, BOTTOM,
		       (int) (event->xkey.state & (ShiftMask | LockMask)),
		       (int) (event->xkey.state & ControlMask));
}

static void
MoveMlinkInput(MlinkWidget w, int x, int direction, int shift, int control)
{
	mlinkCallbackStruct cb;
	int         r;

	if (control)
		ControlMlink(w, direction);
	else if (shift && (direction == TOP || direction == BOTTOM)) {
		MoveShiftCb(w, direction, NORMAL);
		PutMove(direction, 0, 1);
	} else if (CheckSolved(w)) {
		MoveNoTiles(w);
		return;
	} else if (direction == LEFT || direction == RIGHT) {
		SlideMlink(w, direction);
		if (CheckSolved(w)) {
			cb.reason = MLINK_SOLVED;
			XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
		}
	} else {
		if (!PositionTile(w, x, &r))
			return;
		RotateMlink(w, direction, r);
		if (CheckSolved(w)) {
			cb.reason = MLINK_SOLVED;
			XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
		}
	}
}

Boolean
MoveMlink(MlinkWidget w, int direction, int tile, int shift, int fast)
{
	mlinkCallbackStruct cb;

	if (shift && (direction == TOP || direction == BOTTOM)) {
		MoveShiftCb(w, direction, fast);
		PutMove(direction, tile, 1);
		return True;
	} else if (MoveTilesDir(w, direction, tile, False, fast)) {
		cb.reason = MLINK_MOVED;
		XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
		PutMove(direction, tile, 0);
		return True;
	}
	return False;
}

static void
ControlMlink(MlinkWidget w, int direction)
{
	mlinkCallbackStruct cb;

	cb.reason = MLINK_IGNORE;
	switch (direction) {
		case TOP:
			if (w->mlink.faces <= MINFACES)
				return;
			cb.reason = MLINK_DEC_Y;
			break;
		case RIGHT:
			cb.reason = MLINK_INC_X;
			break;
		case BOTTOM:
			if (w->mlink.faces >= MAXFACES)
				return;
			cb.reason = MLINK_INC_Y;
			break;
		case LEFT:
			if (w->mlink.tiles <= MINTILES)
				return;
			cb.reason = MLINK_DEC_X;
			break;
		default:
			{
				char *buf;

				intCat(&buf, "ControlMlink: direction ",
					direction);
				XtWarning(buf);
				free(buf);
			}
	}
	XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
}

static void
RotateMlink(MlinkWidget w, int direction, int tile)
{
	mlinkCallbackStruct cb;

	if (CheckSolved(w)) {
		MoveNoTiles(w);
		return;
	}
	(void) MoveMlink(w, direction, tile, FALSE, NORMAL);
	if (CheckSolved(w)) {
		cb.reason = MLINK_SOLVED;
		XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
	}
	return;
}

static void
SlideMlink(MlinkWidget w, int direction)
{
	mlinkCallbackStruct cb;

	if (CheckSolved(w)) {
		MoveNoTiles(w);
		return;
	}
	if (!MoveMlink(w, direction, 0, FALSE, NORMAL)) {
		cb.reason = MLINK_BLOCKED;
		XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
		return;
	}
	if (CheckSolved(w)) {
		cb.reason = MLINK_SOLVED;
		XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
	}
	return;
}

static int
SelectTiles(MlinkWidget w, int x, int y, int *i, int *j)
{
	*i = (x - w->mlink.delta.x / 2 - w->mlink.puzzleOffset.x) /
		w->mlink.offset.x;
	*j = ((y - w->mlink.delta.y / 2 - w->mlink.puzzleOffset.y) %
	      (w->mlink.faces * w->mlink.offset.y + w->mlink.delta.y - 1)) /
		w->mlink.offset.y;
	if (*i >= 0 && *j >= 0 &&
	    *i < w->mlink.tiles && *j < w->mlink.faces)
		return (*i + w->mlink.tiles * *j -
			w->mlink.spacePosition % w->mlink.tileFaces);
	return -w->mlink.tileFaces;
}

static void
SelectSlideTiles(MlinkWidget w, int pos)
{
	int         n;

	if (pos < 0) {
		if (w->mlink.moveDelay > 0)
			AnimateSlide(w, -pos, RIGHT, NORMAL, True);
		else
			for (n = 1; n <= -pos % w->mlink.tiles; n++) {
				DiscreteMoves(w, w->mlink.spacePosition - 1,
					RIGHT);
		}
	} else if (pos > 0) {
		if (w->mlink.moveDelay > 0)
			AnimateSlide(w, pos, LEFT, NORMAL, True);
		else
			for (n = 1; n <= pos % w->mlink.tiles; n++) {
				DiscreteMoves(w, w->mlink.spacePosition + 1,
					LEFT);
			}
	}
}

#define FACTOR 0.7
#define MAXINTENSITY 0xFFFF

static Pixel
brighter(MlinkWidget w, Pixel pixel)
{
	XColor color;
	int i = (int) ((1 - FACTOR) * MAXINTENSITY);

	color.pixel = pixel;
	XQueryColor(XtDisplay(w), DefaultColormapOfScreen(XtScreen(w)),
		&color);
	if (color.red < i)
		color.red = i;
	if (color.green < i)
		color.green = i;
	if (color.blue < i)
		color.blue = i;
	color.red = (unsigned short) MIN(color.red / FACTOR, MAXINTENSITY);
	color.green = (unsigned short) MIN(color.green / FACTOR, MAXINTENSITY);
	color.blue = (unsigned short) MIN(color.blue / FACTOR, MAXINTENSITY);
	if (XAllocColor(XtDisplay(w), DefaultColormapOfScreen(XtScreen(w)),
			&color))
		return color.pixel;
	return pixel;
}

static Pixel
darker(MlinkWidget w, Pixel pixel)
{
	XColor color;

	color.pixel = pixel;
	XQueryColor(XtDisplay(w), DefaultColormapOfScreen(XtScreen(w)),
		&color);
	color.red = (unsigned short) (color.red * FACTOR);
	color.green = (unsigned short) (color.green * FACTOR);
	color.blue = (unsigned short) (color.blue * FACTOR);
	if (XAllocColor(XtDisplay(w), DefaultColormapOfScreen(XtScreen(w)),
			&color))
		return color.pixel;
	return pixel;
}

static void
SetAllColors(MlinkWidget w, Boolean init)
{
	XGCValues   values;
	XtGCMask    valueMask;
	int         face;

	valueMask = GCForeground | GCBackground;

	if (w->mlink.reverse) {
		values.foreground = w->mlink.foreground;
		values.background = w->mlink.background;
	} else {
		values.foreground = w->mlink.background;
		values.background = w->mlink.foreground;
	}
	if (!init)
		XtReleaseGC((Widget) w, w->mlink.inverseGC);
	w->mlink.inverseGC = XtGetGC((Widget) w, valueMask, &values);
	if (w->mlink.reverse) {
		values.foreground = w->mlink.background;
		values.background = w->mlink.foreground;
	} else {
		values.foreground = w->mlink.foreground;
		values.background = w->mlink.background;
	}
	if (!init)
		XtReleaseGC((Widget) w, w->mlink.puzzleGC);
	w->mlink.puzzleGC = XtGetGC((Widget) w, valueMask, &values);
	if (w->mlink.depth < 2 || w->mlink.mono) {
		if (w->mlink.reverse) {
			values.background = w->mlink.foreground;
			values.foreground = w->mlink.background;
		} else {
			values.foreground = w->mlink.foreground;
			values.background = w->mlink.background;
		}
	} else {
		values.foreground = w->mlink.tileColor;
		values.background = w->mlink.borderColor;
	}
	if (!init)
		XtReleaseGC((Widget) w, w->mlink.tileGC);
	w->mlink.tileGC = XtGetGC((Widget) w, valueMask, &values);
	if (w->mlink.depth >= 2 && !w->mlink.mono) {
		values.foreground = darker(w, w->mlink.tileColor);
	}
	if (!init)
		XtReleaseGC((Widget) w, w->mlink.tileBrighterGC);
	w->mlink.tileBrighterGC = XtGetGC((Widget) w, valueMask, &values);
	if (w->mlink.depth >= 2 && !w->mlink.mono) {
		values.foreground = brighter(w, w->mlink.tileColor);
	}
	if (!init)
		XtReleaseGC((Widget) w, w->mlink.tileDarkerGC);
	w->mlink.tileDarkerGC = XtGetGC((Widget) w, valueMask, &values);
	if (w->mlink.depth < 2 || w->mlink.mono) {
		if (w->mlink.reverse) {
			values.foreground = w->mlink.foreground;
			values.background = w->mlink.background;
		} else {
			values.background = w->mlink.foreground;
			values.foreground = w->mlink.background;
		}
	} else {
		values.foreground = w->mlink.borderColor;
		values.background = w->mlink.tileColor;
	}
	if (!init)
		XtReleaseGC((Widget) w, w->mlink.borderGC);
	w->mlink.borderGC = XtGetGC((Widget) w, valueMask, &values);
	for (face = 0; face < MAXFACES; face++)
		GetColor(w, face, init);
}

static void
GetColor(MlinkWidget w, int face, Boolean init)
{
	XGCValues   values;
	XtGCMask    valueMask;
	XColor      colorCell, rgb;

	valueMask = GCForeground | GCBackground;
	if (w->mlink.reverse) {
		values.background = w->mlink.foreground;
	} else {
		values.background = w->mlink.background;
	}
	if (w->mlink.depth > 1 && !w->mlink.mono) {
		if (XAllocNamedColor(XtDisplay(w),
				  DefaultColormapOfScreen(XtScreen(w)),
				w->mlink.faceName[face], &colorCell, &rgb)) {
			values.foreground = w->mlink.faceColor[face] = colorCell.pixel;
			if (!init)
				XtReleaseGC((Widget) w, w->mlink.faceGC[face]);
			w->mlink.faceGC[face] = XtGetGC((Widget) w, valueMask, &values);
			return;
		} else {
			char *buf1, *buf2;
 
			stringCat(&buf1, "Color name \"",
				w->mlink.faceName[face]);
			stringCat(&buf2, buf1, "\" is not defined for face ");
			free(buf1);
			intCat(&buf1, buf2, face);
			free(buf2);
			XtWarning(buf1);
			free(buf1);
		}
	}
	if (w->mlink.reverse) {
		values.background = w->mlink.background;
		values.foreground = w->mlink.foreground;
	} else {
		values.background = w->mlink.foreground;
		values.foreground = w->mlink.background;
	}
	if (!init)
		XtReleaseGC((Widget) w, w->mlink.faceGC[face]);
	w->mlink.faceGC[face] = XtGetGC((Widget) w, valueMask, &values);
}

static void
MoveShiftCb(MlinkWidget w, int direction, int fast)
{
	mlinkCallbackStruct cb;

	(void) MoveTilesDir(w, direction, 0, True, fast);
	cb.reason = MLINK_CONTROL;
	XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
}

static void
CheckTiles(MlinkWidget w)
{
	char *buf1 = NULL, *buf2 = NULL;

	if (w->mlink.tiles < MINTILES) {
		intCat(&buf1,
			"Number of tiles in X direction out of bounds, use at least ",
			MINTILES);
		stringCat(&buf2, buf1, ", defaulting to ");
		free(buf1);
		intCat(&buf1, buf2, DEFAULTTILES);
		free(buf2);
		XtWarning(buf1);
		free(buf1);
		w->mlink.tiles = DEFAULTTILES;
	}
	if (w->mlink.faces < MINFACES || w->mlink.faces > MAXFACES) {
		intCat(&buf1,
			"Number of faces out of bounds, use ",
			MINFACES);
		stringCat(&buf2, buf1, "..");
		free(buf1);
		intCat(&buf1, buf2, MAXFACES);
		free(buf2);
		stringCat(&buf2, buf1, ", defaulting to ");
		free(buf1);
		intCat(&buf1, buf2, DEFAULTFACES);
		free(buf2);
		XtWarning(buf1);
		free(buf1);
		w->mlink.faces = DEFAULTFACES;
	}
	if (w->mlink.base < MINBASE  || w->mlink.base > MAXBASE) {
		intCat(&buf1, "Base out of bounds, use ", MINBASE);
		stringCat(&buf2, buf1, "..");
		free(buf1);
		intCat(&buf1, buf2, MAXBASE);
		free(buf2);
		stringCat(&buf2, buf1, ", defaulting to ");
		free(buf1);
		intCat(&buf1, buf2, DEFAULTBASE);
		free(buf2);
		XtWarning(buf1);
		free(buf1);
		w->mlink.base = DEFAULTBASE;
	}
}

static void
ResetTiles(MlinkWidget w)
{
	int         i;

	w->mlink.tileFaces = w->mlink.tiles * w->mlink.faces;
	if (w->mlink.tileOfPosition)
		free(w->mlink.tileOfPosition);
	if (!(w->mlink.tileOfPosition = (int *)
	      malloc(sizeof (int) * w->mlink.tileFaces)))
		            XtError("Not enough memory, exiting.");

	if (startPosition)
		free(startPosition);
	if (!(startPosition = (int *)
	      malloc(sizeof (int) * w->mlink.tileFaces)))
		            XtError("Not enough memory, exiting.");

	w->mlink.spacePosition = w->mlink.tileFaces - 1;
	w->mlink.tileOfPosition[w->mlink.tileFaces - 1] = 0;
	for (i = 1; i < w->mlink.tileFaces; i++)
		w->mlink.tileOfPosition[i - 1] = i;
	FlushMoves(w);
	w->mlink.currentRef = -w->mlink.tileFaces;
	w->mlink.started = False;
}

static void
ResizeTiles(MlinkWidget w)
{
	w->mlink.digitOffset.x = 3;
	w->mlink.digitOffset.y = 4;
}

static void
MoveNoTiles(MlinkWidget w)
{
	mlinkCallbackStruct cb;

	cb.reason = MLINK_IGNORE;
	XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
}

static      Boolean
MoveTilesDir(MlinkWidget w, int direction, int tile, Boolean all, int fast)
{
	int start, finish;

	start = tile;
	finish = tile;
	if (all) {
		start = 0;
		finish = w->mlink.tiles - 1;
	} else if (w->mlink.middle && tile > 0 && tile < w->mlink.tiles - 1) {
		start = 1;
		finish =  w->mlink.tiles - 2;
	}
	switch (direction) {
	case TOP:
		RotateTiles(w, start, finish, TOP, fast);
		return True;
	case RIGHT:
		if (Column(w, w->mlink.spacePosition) > 0) {
			MoveTiles(w, w->mlink.spacePosition - 1, fast);
			return True;
		}
		break;
	case BOTTOM:
		RotateTiles(w, start, finish, BOTTOM, fast);
		return True;
	case LEFT:
		if (Column(w, w->mlink.spacePosition) < w->mlink.tiles - 1) {
			MoveTiles(w, w->mlink.spacePosition + 1, fast);
			return True;
		}
		break;
	default:
		{
			char *buf;

			intCat(&buf, "MoveTilesDir: direction ", direction);
			XtWarning(buf);
			free(buf);
		}
	}
	return False;
}

static void
GetTiles(MlinkWidget w)
{
	FILE       *fp;
	int         c, i, tiles, faces, middle, orient, moves;
	mlinkCallbackStruct cb;
	char *buf1 = NULL, *buf2 = NULL;
	char *fname, *lname, *name;

	stringCat(&buf1, CURRENTDELIM, LOGFILE);
	lname = buf1;
	stringCat(&buf1, LOGPATH, FINALDELIM);
	stringCat(&buf2, buf1, LOGFILE);
	free(buf1);
	fname = buf2;
	/* Try current directory first. */
	name = lname;
	if ((fp = fopen(name, "r")) == NULL) {
		name = fname;
		if ((fp = fopen(name, "r")) == NULL) {
			stringCat(&buf1, "Can not read (get) ", lname);
			stringCat(&buf2, buf1, " or ");
			free(buf1);
			stringCat(&buf1, buf2, fname);
			free(buf2);
			XtWarning(buf1);
			free(buf1);
			free(lname);
			free(fname);
			return;
		}
/* Probably annoying */
#if 0
		else {
			stringCat(&buf1, "Can not read (get) ", lname);
			stringCat(&buf2, buf1, ", falling back to ");
			free(buf1);
			stringCat(&buf1, buf2, fname);
			free(buf2);
			XtWarning(buf1);
			free(buf1);
		}
#endif
	}
	FlushMoves(w);
	while ((c = getc(fp)) != EOF && c != SYMBOL);
	(void) fscanf(fp, "%d", &tiles);
	if (tiles >= MINTILES) {
		for (i = w->mlink.tiles; i < tiles; i++) {
			cb.reason = MLINK_INC_X;
			XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
		}
		for (i = w->mlink.tiles; i > tiles; i--) {
			cb.reason = MLINK_DEC_X;
			XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
		}
	} else {
		stringCat(&buf1, name, " corrupted: tiles ");
		intCat(&buf2, buf1, tiles);
		free(buf1);
		stringCat(&buf1, buf2, " should be between ");
		free(buf2);
		intCat(&buf2, buf1, MINTILES);
		free(buf1);
		stringCat(&buf1, buf2, " and MAXINT");
		free(buf2);
		XtWarning(buf1);
		free(buf1);
	}
	while ((c = getc(fp)) != EOF && c != SYMBOL);
	(void) fscanf(fp, "%d", &faces);
	if (faces >= MINFACES && faces <= MAXFACES) {
		for (i = w->mlink.faces; i < faces; i++) {
			cb.reason = MLINK_INC_Y;
			XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
		}
		for (i = w->mlink.faces; i > faces; i--) {
			cb.reason = MLINK_DEC_Y;
			XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
		}
	} else {
		stringCat(&buf1, name, " corrupted: faces ");
		intCat(&buf2, buf1, faces);
		free(buf1);
		stringCat(&buf1, buf2, " should be between ");
		free(buf2);
		intCat(&buf2, buf1, MINFACES);
		free(buf1);
		stringCat(&buf1, buf2, " and ");
		free(buf2);
		intCat(&buf2, buf1, MAXFACES);
		free(buf1);
		XtWarning(buf2);
		free(buf2);
	}
	while ((c = getc(fp)) != EOF && c != SYMBOL);
	(void) fscanf(fp, "%d", &middle);
	if (w->mlink.middle != (Boolean) middle) {
		cb.reason = MLINK_MIDDLE;
		XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
	}
	while ((c = getc(fp)) != EOF && c != SYMBOL);
	(void) fscanf(fp, "%d", &orient);
	if (w->mlink.orient != (Boolean) orient) {
		cb.reason = MLINK_ORIENT;
		XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
	}
	while ((c = getc(fp)) != EOF && c != SYMBOL);
	(void) fscanf(fp, "%d", &moves);
	ScanStartPosition(fp, w);
	cb.reason = MLINK_RESTORE;
	XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
	SetStartPosition(w);
	ScanMoves(fp, w, moves);
	(void) fclose(fp);
	(void) printf("%s: tiles %d, faces %d, middle %d, orient %d, moves %d.\n",
		      name, tiles, faces, middle, orient, moves);
	free(lname);
	free(fname);
	w->mlink.cheat = True; /* Assume the worst. */
}

static void
WriteTiles(MlinkWidget w)
{
	FILE       *fp;
	char *buf1 = NULL, *buf2 = NULL;
	char *fname, *lname, *name;

	stringCat(&buf1, CURRENTDELIM, LOGFILE);
	lname = buf1;
	stringCat(&buf1, LOGPATH, FINALDELIM);
	stringCat(&buf2, buf1, LOGFILE);
	free(buf1);
	fname = buf2;
	/* Try current directory first. */
	name = lname;
	if ((fp = fopen(name, "w")) == NULL) {
		name = fname;
		if ((fp = fopen(name, "w")) == NULL) {
			stringCat(&buf1, "Can not write to ", lname);
			stringCat(&buf2, buf1, " or ");
			free(buf1);
			stringCat(&buf1, buf2, fname);
			free(buf2);
			XtWarning(buf1);
			free(buf1);
			free(lname);
			free(fname);
			return;
		}
/* Probably annoying */
#if 0
		else {
			stringCat(&buf1, "Can not write to ", lname);
			stringCat(&buf2, buf1, ", falling back to ");
			free(buf1);
			stringCat(&buf1, buf2, fname);
			free(buf2);
			XtWarning(buf1);
			free(buf1);
		}
#endif
	}
	(void) fprintf(fp, "tiles%c %d\n", SYMBOL, w->mlink.tiles);
	(void) fprintf(fp, "faces%c %d\n", SYMBOL, w->mlink.faces);
	(void) fprintf(fp, "middle%c %d\n", SYMBOL,
		(w->mlink.middle) ? 1 : 0);
	(void) fprintf(fp, "orient%c %d\n", SYMBOL,
		(w->mlink.orient) ? 1 : 0);
	(void) fprintf(fp, "moves%c %d\n", SYMBOL, NumMoves());
	PrintStartPosition(fp, w);
	PrintMoves(fp);
	(void) fclose(fp);
	(void) printf("Saved to %s.\n", name);
	free(lname);
	free(fname);
}

static void
ClearTiles(MlinkWidget w)
{
	mlinkCallbackStruct cb;

	if (w->mlink.currentRef != -w->mlink.tileFaces)
		return;
	ResetTiles(w);
	DrawAllTiles(w);
	cb.reason = MLINK_RESET;
	XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
}

static void
UndoTiles(MlinkWidget w)
{
	if (MadeMoves() && w->mlink.currentRef == -w->mlink.tileFaces) {
		int         direction, tile, shift;
		mlinkCallbackStruct cb;

		GetMove(&direction, &tile, &shift);
		direction = (direction + COORD / 2) % COORD;
		if (shift && (direction == TOP || direction == BOTTOM)) {
			MoveShiftCb(w, direction, DOUBLE);
		} else if (MoveTilesDir(w, direction, tile, False, DOUBLE)) {
			cb.reason = MLINK_UNDO;
			XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
		} else {
			char *buf1, *buf2;

			intCat(&buf1, "Move ", direction);
			stringCat(&buf2, buf1, " can not be made");
			free(buf1);
			XtWarning(buf2);
			free(buf2);
		}
	}
}

static void
RandomizeTiles(MlinkWidget w)
{
	mlinkCallbackStruct cb;

	if (w->mlink.currentRef != -w->mlink.tileFaces)
		return;
	w->mlink.cheat = False;
	/* First interchange tiles an even number of times */
	if (w->mlink.tiles > 1 && w->mlink.faces > 1 && w->mlink.tileFaces > 4) {
		int         pos, count = 0;

		for (pos = 0; pos < w->mlink.tileFaces; pos++) {
			int         randomPos = pos;

			while (randomPos == pos)
				randomPos = NRAND(w->mlink.tileFaces);
			count += ExchangeTiles(w, pos, randomPos);
		}
		if (count % 2 &&
		    !ExchangeTiles(w, 0, 1) &&
		    !ExchangeTiles(w, w->mlink.tileFaces - 2,
		               w->mlink.tileFaces - 1)) {
			XtWarning("RandomizeTiles: should not get here");
		}
		DrawAllTiles(w);
		XFlush(XtDisplay(w));
	}
	/* randomly position space */
	/* Idea for this came from "puzzle" by Don Bennett, HP Labs */
	{
		int n, s, e, c;

		/* order matters */
		s = Row(w, w->mlink.spacePosition);
		c = Column(w, w->mlink.spacePosition);
		e = NRAND(w->mlink.faces);
		for (n = 0; n < e - s; n++)
			(void) MoveTilesDir(w, BOTTOM, c, False, INSTANT);
		for (n = 0; n < s - e; n++)
			(void) MoveTilesDir(w, TOP, c, False, INSTANT);
		s = Column(w, w->mlink.spacePosition);
		e = NRAND(w->mlink.tiles);
		for (n = 0; n < e - s; n++)
			(void) MoveTilesDir(w, LEFT, 0, False, INSTANT);
		for (n = 0; n < s - e; n++)
			(void) MoveTilesDir(w, RIGHT, 0, False, INSTANT);
		FlushMoves(w);
		cb.reason = MLINK_RANDOMIZE;
		XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
	}
#if 0
	/* Now move the space around randomly */
	if (w->mlink.tiles > 1 || w->mlink.faces > 1) {
		int         big = w->mlink.tileFaces + NRAND(2);
		int         lastDirection = -1;
		int         randomDirection;

		cb.reason = MLINK_RESET;
		XtCallCallbacks((Widget) w, XtNselectCallback, &cb);

#ifdef DEBUG
		big = 3;
#endif
		if (big > 100)
			big = 100;
		while (big--) {
			randomDirection = NRAND(COORD);

#ifdef DEBUG
			sleep(1);
#endif

			if ((randomDirection + COORD / 2) % COORD != lastDirection ||
			    w->mlink.tiles == 1 || w->mlink.faces == 1) {
				if (MoveMlink(w, randomDirection, NRAND(w->mlink.tiles), FALSE, INSTANT))
					lastDirection = randomDirection;
				else
					big++;
			}
		}
		FlushMoves(w);
		cb.reason = MLINK_RANDOMIZE;
		XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
	}
#endif
	if (CheckSolved(w)) {
		cb.reason = MLINK_SOLVED;
		XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
	}
}

static void
SolveTiles(MlinkWidget w)
{
	if (CheckSolved(w) || w->mlink.currentRef != -w->mlink.tileFaces)
		return;
	{
		mlinkCallbackStruct cb;

		cb.reason = MLINK_SOLVE_MESSAGE;
		XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
	}
}

static void
OrientizeTiles(MlinkWidget w)
{
	mlinkCallbackStruct cb;

	cb.reason = MLINK_ORIENT;
	XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
}

static void
MiddleTiles(MlinkWidget w)
{
	mlinkCallbackStruct cb;

	cb.reason = MLINK_MIDDLE;
	XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
}

static void
MoveTiles(MlinkWidget w, int from, int fast)
{
	int         tempTile;

	if (fast != INSTANT && w->mlink.moveDelay > 0) {
		int i =	w->mlink.currentTile = Column(w, from);
		int j = w->mlink.currentFace = Row(w, from);
		int l = i + w->mlink.tiles * j - w->mlink.spacePosition;
		w->mlink.currentRef = from;

		if (l / w->mlink.tiles == 0 &&
				j == Row(w, w->mlink.spacePosition)) {
			if (l < 0) {
				AnimateSlide(w, -l, RIGHT, fast, False);
			} else {        /* (l > 0) */
				AnimateSlide(w, l, LEFT, fast, False);
			}
		}
		w->mlink.currentRef = -w->mlink.tileFaces;
	} else {
		tempTile = w->mlink.tileOfPosition[from];
		w->mlink.tileOfPosition[from] =
			w->mlink.tileOfPosition[w->mlink.spacePosition];
		w->mlink.tileOfPosition[w->mlink.spacePosition] = tempTile;
		DrawTile(w, w->mlink.spacePosition, False, False, FALSE, 0, 0);
		w->mlink.spacePosition = from;
		DrawTile(w, w->mlink.spacePosition, True, True, FALSE, 0, 0);
	}
}

static void
AnimateRotate(MlinkWidget w, int start, int finish, int direction, int fast)
{
	Pixmap bufferBox;
	XWindowAttributes xgwa;
	int size = finish - start + 1;
	int moveI, gapJ, moveJ, inc, i;

	moveI = w->mlink.tileSize.x + w->mlink.delta.x;
	gapJ = w->mlink.tileSize.y * fast / w->mlink.numSlices;
	moveJ = w->mlink.tileSize.y + w->mlink.delta.y + 1;
	if (gapJ == 0)
		gapJ++;
	
	(void) XGetWindowAttributes(XtDisplay(w), XtWindow(w), &xgwa);

	if ((bufferBox = XCreatePixmap(XtDisplay(w), XtWindow(w),
       			size * moveI + 1,
			w->mlink.puzzleSize.y,
			xgwa.depth)) == None) {
		XtError("Not enough memory, exiting.");
	}
	XCopyArea(XtDisplay(w), XtWindow(w), bufferBox,
		w->mlink.puzzleGC,
		start * w->mlink.offset.x + 1,
		w->mlink.puzzleOffset.y,
		size * moveI + 1,
		w->mlink.puzzleSize.y,
		0, 0);
	XSetGraphicsExposures(XtDisplay(w), w->mlink.puzzleGC, False);
	for (inc = 0; inc < moveJ + gapJ; inc += gapJ) {
		if (inc > moveJ)
			inc = moveJ;
		if (direction == TOP)
			i = inc;
		else
			i = w->mlink.puzzleSize.y - inc - w->mlink.puzzleOffset.y + 1;
		XCopyArea(XtDisplay(w), bufferBox, XtWindow(w),
			w->mlink.puzzleGC,
			0, i,
			size * moveI + 1,
			w->mlink.puzzleSize.y - i,
			start * w->mlink.offset.x + 1,
			w->mlink.puzzleOffset.y + 1);
		XCopyArea(XtDisplay(w), bufferBox, XtWindow(w),
			w->mlink.puzzleGC,
			0, 0,
			size * moveI,
			i - w->mlink.delta.y + w->mlink.puzzleOffset.y + 1,
			start * w->mlink.offset.x + 1,
			w->mlink.puzzleSize.y - i + w->mlink.puzzleOffset.y - w->mlink.delta.y + 1);
		XFlush(XtDisplay(w));
		Sleep(w->mlink.moveDelay / fast);
	}
	XFreePixmap(XtDisplay(w), bufferBox);
}

static void
RotateTiles(MlinkWidget w, int start, int finish, int direction, int fast)
{
	int         newR, i, k, tempTile, currTile, pos, newPos;
	int         r = 0, space = -1;
	Boolean	    animate = (fast != INSTANT && w->mlink.moveDelay > 0);

	if (animate)
		AnimateRotate(w, start, finish, direction, fast);
	for  (i = start; i <= finish; i++) {
		currTile = w->mlink.tileOfPosition[r * w->mlink.tiles + i];
		for (k = 1; k <= w->mlink.faces; k++) {
			newR = (direction == TOP) ?  (r +
				w->mlink.faces - 1) % w->mlink.faces :
				(r + 1) % w->mlink.faces;
			pos = r * w->mlink.tiles + i;
			newPos = newR * w->mlink.tiles + i;
			tempTile = w->mlink.tileOfPosition[newPos];
			w->mlink.tileOfPosition[newPos] = currTile;
			currTile = tempTile;
			if (w->mlink.spacePosition == pos) {
				if (!animate)
					DrawTile(w, newPos, True, True,
						FALSE, 0, 0);
				space = newPos;
			} else if (!animate)
				DrawTile(w, newPos, False, False,
					FALSE, 0, 0);
			r = newR;
		}
		if (space >= 0)
			w->mlink.spacePosition = space;
	}
	if (animate) {
		/* Would be nice if this was not needed at all but I did
		   try to minimize the redraw */
		int moveI = w->mlink.tileSize.x + w->mlink.delta.x;
		
		XFillRectangle(XtDisplay(w), XtWindow(w), w->mlink.inverseGC,
			start * moveI + w->mlink.puzzleOffset.x + 1, 0,
			(finish - start + 1) * moveI + 1, w->core.height);
		DrawFrame(w, (w->mlink.focus) ? w->mlink.puzzleGC :
			w->mlink.inverseGC);
		DrawAllTilesForColumns(w, start, finish);
	}
}

static int
ExchangeTiles(MlinkWidget w, int pos1, int pos2)
{
	int         tempTile;

	if (w->mlink.tileOfPosition[pos1] <= 0)
		return FALSE;
	else if (w->mlink.tileOfPosition[pos2] <= 0)
		return FALSE;
	tempTile = w->mlink.tileOfPosition[pos1];
	w->mlink.tileOfPosition[pos1] = w->mlink.tileOfPosition[pos2];
	w->mlink.tileOfPosition[pos2] = tempTile;
	return TRUE;
}

static int
TileNFromSpace(MlinkWidget w, int n, int direction)
{
	int pos = w->mlink.spacePosition;

	if (direction == LEFT || direction == RIGHT)
		return (pos + ((direction == RIGHT) ? -n : n));
	else /* direction == TOP || direction == BOTTOM */
		return (pos + (w->mlink.tiles * ((direction == BOTTOM) ?
			-n : n)));
}

static void
DrawFrame(MlinkWidget w, GC gc)
{
	int         sumX, sumY, offsetX, offsetY, r;

	offsetX = w->mlink.puzzleOffset.x;
	offsetY = w->mlink.puzzleOffset.y;
	sumX = w->mlink.tiles * w->mlink.offset.x + w->mlink.delta.x / 2 + 1;
	sumY = w->mlink.offset.y;
	XFillRectangle(XtDisplay(w), XtWindow(w), w->mlink.borderGC,
		0, 0, w->core.width, offsetY);
	XFillRectangle(XtDisplay(w), XtWindow(w), w->mlink.borderGC,
		0, w->core.height - offsetY - 1,
		w->core.width, offsetY + 2);
	XFillRectangle(XtDisplay(w), XtWindow(w), w->mlink.borderGC,
		0, 0, offsetX, w->core.height);
	XFillRectangle(XtDisplay(w), XtWindow(w), w->mlink.borderGC,
		offsetX + sumX, 0,
		w->core.width - sumX - offsetX, w->core.height);
	XFillRectangle(XtDisplay(w), XtWindow(w), gc,
		offsetX, offsetY, sumX, 1);
	XFillRectangle(XtDisplay(w), XtWindow(w), gc,
		offsetX, offsetY, 1, w->core.height - 2 * offsetY - 1);
	XFillRectangle(XtDisplay(w), XtWindow(w), gc,
		sumX + offsetX, offsetY, 1, w->core.height - 2 * offsetY - 1);
	for (r = 1; r < w->mlink.faces; r++) {
		offsetY += sumY;
		XFillRectangle(XtDisplay(w), XtWindow(w), gc,
			       offsetX, offsetY, sumX, 1);
	}
	XFillRectangle(XtDisplay(w), XtWindow(w), gc,
	       offsetX, offsetY + sumY, sumX + 1, 1);
}

void
DrawAllTiles(MlinkWidget w)
{
	int         k;

	for (k = 0; k < w->mlink.tileFaces; k++)
		DrawTile(w, k, (w->mlink.tileOfPosition[k] <= 0),
			 (w->mlink.tileOfPosition[k] <= 0), FALSE, 0, 0);
}

static void
DrawAllTilesForColumns(MlinkWidget w, int start, int finish)
{
	int         i, j, k;

	for (i = start; i <= finish; i++) {
		for (j = 0; j < w->mlink.faces; j++) {
			k = j * w->mlink.tiles + i;
			if (w->mlink.tileOfPosition[k] > 0)
			  DrawTile(w, k,
			    (w->mlink.tileOfPosition[k * w->mlink.tiles + i] <= 0),
			    (w->mlink.tileOfPosition[k] <= 0), FALSE, 0, 0);
		}
	}
}

static void
DrawTile(MlinkWidget w, int pos, Boolean blank, Boolean erase,
		int pressedOffset, int offsetX, int offsetY)
{
	int         dx, dy, tile = w->mlink.tileOfPosition[pos];
	GC          linkGC, tileGC, borderGC;

	if (erase) {
		tileGC = w->mlink.inverseGC;
		borderGC = w->mlink.inverseGC;
	} else if (pressedOffset) {
		tileGC = w->mlink.borderGC;
		borderGC = w->mlink.tileGC;
	} else {
		tileGC = w->mlink.tileGC;
		borderGC = w->mlink.borderGC;
	}
	if ((w->mlink.depth < 2 || w->mlink.mono) && pressedOffset) {
		linkGC = borderGC;
	} else {
		int         face = ((tile - 1) / w->mlink.tiles + 1) % w->mlink.faces;

		linkGC = w->mlink.faceGC[face];
		if (w->mlink.depth >= 2 && !w->mlink.mono) {
			if (w->mlink.faceColor[face] == w->mlink.tileColor)  {
				linkGC = (pressedOffset) ? tileGC : borderGC;
			}
		}
	}
	dx = Column(w, pos) * w->mlink.offset.x + w->mlink.delta.x +
		w->mlink.puzzleOffset.x + pressedOffset - 1;
	dy = Row(w, pos) * w->mlink.offset.y + w->mlink.delta.y +
		w->mlink.puzzleOffset.y + pressedOffset - 1;
	dx += offsetX;
	dy += offsetY;
	if (blank)
		XFillRectangle(XtDisplay(w), XtWindow(w), tileGC, dx, dy,
		       w->mlink.tileSize.x + 1, w->mlink.tileSize.y - 1);

	else
		fill3DRect(w, tileGC, w->mlink.tileDarkerGC, w->mlink.tileBrighterGC,
                        dx, dy, w->mlink.tileSize.x + 1, w->mlink.tileSize.y - 1,
                        pressedOffset == 0);
		/* XFillRectangle(XtDisplay(w), XtWindow(w), tileGC,
		       dx + 1, dy + 1,
		       w->mlink.tileSize.x, w->mlink.tileSize.y - 2); */
	if (!blank) {
		int         i = 0, digitX = 0, t = tile;
		char        buf[66];

		buf[0] = '\0';
		if (w->mlink.orient || w->mlink.depth < 2 || w->mlink.mono) {
			if (w->mlink.depth < 2 || w->mlink.mono) {
				if (w->mlink.orient) {
					int         last;

					last = int2String(buf, tile, w->mlink.base, False);
					buf[last] = w->mlink.faceName[((tile - 1) / w->mlink.tiles + 1) %
							  w->mlink.faces][0];
					buf[last + 1] = '\0';
					t *= w->mlink.base;
				} else {
					buf[0] = w->mlink.faceName[((tile - 1) /
								    w->mlink.tiles + 1) % w->mlink.faces][0];
					buf[1] = '\0';
					t = 1;
				}
			} else {
				(void) int2String(buf, tile, w->mlink.base, True);
			}
			while (t >= 1) {
				t /= w->mlink.base;
				digitX += w->mlink.digitOffset.x;
				i++;
			}
			(void) XDrawString(XtDisplay(w), XtWindow(w), linkGC,
				    dx + w->mlink.tileSize.x / 2 - digitX,
				    dy + w->mlink.tileSize.y / 2 + w->mlink.digitOffset.y, buf, i);
		}
	}
	if (tile != 0 &&
	    ((tile != w->mlink.tileFaces - 1 && w->mlink.tiles > 1) ||
	     w->mlink.tiles > 2))
		DrawLink(w, linkGC, pos, offsetX + pressedOffset,
			offsetY + pressedOffset);
	XDrawRectangle(XtDisplay(w), XtWindow(w), borderGC, dx, dy,
		       w->mlink.tileSize.x + 1, w->mlink.tileSize.y - 1);
}

#define MULT 64
#define THICKNESS 7
static void
DrawLink(MlinkWidget w, GC gc, int pos, int offsetX, int offsetY)
{
	int         dx, dy, sizex, sizey, tile = w->mlink.tileOfPosition[pos],
	            thick;

	sizex = w->mlink.offset.x * 3 / 2 - 3;
	sizey = w->mlink.offset.y * 3 / 4 - 3;
	dx = Column(w, pos) * w->mlink.offset.x + w->mlink.delta.x +
		w->mlink.puzzleOffset.x - sizex / 2 + offsetX;
	dy = Row(w, pos) * w->mlink.offset.y + w->mlink.delta.y +
		w->mlink.puzzleOffset.y + w->mlink.tileSize.y / 2 - sizey / 2 -
		1 + offsetY;
	thick = MIN(sizey, sizex) / 3;
	thick = MIN(thick, THICKNESS);
	if (thick > 1) {
		XSetLineAttributes(XtDisplay(w), gc, thick,
				   LineSolid, CapNotLast, JoinRound);
		if (tile % w->mlink.tiles == 0 || tile == w->mlink.tileFaces - 1)
			XDrawArc(XtDisplay(w), XtWindow(w), gc, dx, dy,
				 sizex, sizey, 89 * MULT, -179 * MULT);
		else if (tile % w->mlink.tiles == 1)
			XDrawArc(XtDisplay(w), XtWindow(w), gc, dx + w->mlink.tileSize.x - 1, dy,
				 sizex, sizey, 90 * MULT, 180 * MULT);
		else {
			XDrawArc(XtDisplay(w), XtWindow(w), gc, dx, dy,
				 sizex, sizey, 89 * MULT, -124 * MULT);
			XDrawArc(XtDisplay(w), XtWindow(w), gc, dx, dy,
				 sizex, sizey, -90 * MULT, 30 * MULT);
			dx += w->mlink.tileSize.x - 1;
			XDrawArc(XtDisplay(w), XtWindow(w), gc, dx, dy,
				 sizex, sizey, 90 * MULT, 30 * MULT);
			XDrawArc(XtDisplay(w), XtWindow(w), gc, dx, dy,
				 sizex, sizey, -90 * MULT, -125 * MULT);
		}
		XSetLineAttributes(XtDisplay(w), gc, 1, LineSolid, CapNotLast, JoinRound);
	}
}

#if 0
#define MULT 64
#define THICKNESS 8
static void
DrawLink(MlinkWidget w, GC gc, int pos, int offsetX, int offsetY)
{
	int         dx, dy, sizex, sizey, i, tile = w->mlink.tileOfPosition[pos];

	for (i = 0; i < THICKNESS; i++) {
		sizex = w->mlink.offset.x * 3 / 2 - i;
		sizey = w->mlink.offset.y * 3 / 4 - i;
		dx = Column(w, pos) * w->mlink.offset.x + w->mlink.delta.x +
			w->mlink.puzzleOffset.x - sizex / 2 + offsetX;
		dy = Row(w, pos) * w->mlink.offset.y + w->mlink.delta.y +
			w->mlink.puzzleOffset.y + w->mlink.tileSize.y / 2 -
			sizey / 2 + offsetY;
		if (tile % w->mlink.tiles == 0 || tile == w->mlink.tileFaces - 1)
			XDrawArc(XtDisplay(w), XtWindow(w), gc, dx, dy,
				 sizex, sizey, 89 * MULT, -179 * MULT);
		else if (tile % w->mlink.tiles == 1)
			XDrawArc(XtDisplay(w), XtWindow(w), gc, dx + w->mlink.tileSize.x - 1, dy,
				 sizex, sizey, 90 * MULT, 180 * MULT);
		else {
			XDrawArc(XtDisplay(w), XtWindow(w), gc, dx, dy,
				 sizex, sizey, 89 * MULT, -32 * MULT);
			XDrawArc(XtDisplay(w), XtWindow(w), gc, dx, dy,
				 sizex, sizey, -90 * MULT, 128 * MULT);
			dx += w->mlink.tileSize.x - 1;
			XDrawArc(XtDisplay(w), XtWindow(w), gc, dx, dy,
				 sizex, sizey, 90 * MULT, 128 * MULT);
			XDrawArc(XtDisplay(w), XtWindow(w), gc, dx, dy,
				 sizex, sizey, -90 * MULT, -33 * MULT);
		}
	}
}
#endif

static void
fill3DRect(MlinkWidget w, GC gc, GC darkerGC, GC brighterGC,
         int x, int y, int width, int height, Boolean raised)
{
        GC currentGC = (raised) ? gc : darkerGC;

        if (width > 2 && height > 2)
                XFillRectangle(XtDisplay(w), XtWindow(w), currentGC,
                        x + 1, y + 1, width - 2, height - 2);
        currentGC = (raised) ? brighterGC : darkerGC;
        XFillRectangle(XtDisplay(w), XtWindow(w), currentGC,
                x, y, 1, height);
        if (width > 1)
                XFillRectangle(XtDisplay(w), XtWindow(w), currentGC,
                        x + 1, y, width - 2, 1);
        currentGC = (raised) ? darkerGC : gc;
        if (width > 1 && height > 1)
                XFillRectangle(XtDisplay(w), XtWindow(w), currentGC,
                        x + 1, y + height - 1, width - 1, 1);
        if (width > 1 && height > 1)
                XFillRectangle(XtDisplay(w), XtWindow(w), currentGC,
                        x + width - 1, y, 1, height - 1);
}

void
AnimateSlide(MlinkWidget w, int numTiles, int dir, int fast, Boolean logMoves)
{
	mlinkCallbackStruct cb;
	int inc, aTile;
	int fillBeginPos, fillBeginI = 0, gapI, moveI;
	int dx, dy, pos, posNext;
	int ix = 0, iy = 0;

	fillBeginPos = TileNFromSpace(w, numTiles, dir);
	fillBeginI = Column(w, fillBeginPos) * w->mlink.offset.x +
	  w->mlink.delta.x + w->mlink.puzzleOffset.x;
	gapI = w->mlink.tileSize.x * fast / w->mlink.numSlices;
	moveI = w->mlink.tileSize.x + w->mlink.delta.x; 
	if (gapI == 0)
	  gapI++;
	if (numTiles < 0)
	  numTiles = -numTiles;
	for (inc = 0; inc < moveI + gapI; inc += gapI) {
	  if (inc > moveI)
	    inc = moveI;
	  for (aTile = 0; aTile < numTiles; aTile++) {
	    if (logMoves && inc == 0) {
	      cb.reason = MLINK_MOVED;
	      XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
	      PutMove(dir, 0, 0);
	    }
	    posNext = TileNFromSpace(w, aTile + 1, dir);
	    /* Calculate deltas */
	    dx = Column(w, posNext) * w->mlink.offset.x + w->mlink.delta.x +
	      w->mlink.puzzleOffset.x;
	    dy = Row(w, posNext) * w->mlink.offset.y + w->mlink.delta.y +
	      w->mlink.puzzleOffset.y;
	    ix = ((dir == RIGHT) ? inc : -inc);
	    iy = 0;
	    DrawTile(w, posNext, False, False, FALSE, ix, iy);
	    /* Erase old slivers */
	    ix += dx;
	    iy += dy;
	    if (aTile < numTiles - 1) {
	      switch (dir) {
	        case RIGHT:
	          if (gapI - 1 > 0)
	          XFillRectangle(XtDisplay(w), XtWindow(w),
	            w->mlink.inverseGC,
	            ix - gapI, iy - 1, gapI - 1, w->mlink.tileSize.y + 1);
	          break;
	        case LEFT:
	          if (gapI - 1 > 0)
	          XFillRectangle(XtDisplay(w), XtWindow(w),
	            w->mlink.inverseGC,
	            ix + w->mlink.tileSize.x + 1, iy - 1,
	            gapI - 1, w->mlink.tileSize.y + 1);
		  break;
	      }
	    } else {
	      switch (dir) {
	        case RIGHT:
	          if (ix - fillBeginI - 1 > 0)
	          XFillRectangle(XtDisplay(w), XtWindow(w),
	            w->mlink.inverseGC,
	            fillBeginI - 1, iy - 1,
	            ix - fillBeginI, w->mlink.tileSize.y + 1);
	          break;
	        case LEFT:
	          XFillRectangle(XtDisplay(w), XtWindow(w),
	            w->mlink.inverseGC,
	            ix + w->mlink.tileSize.x + 1, iy - 1,
	            fillBeginI - ix, w->mlink.tileSize.y + 1);
		 break;
	      }
	    }
	  }
	  XFlush(XtDisplay(w));
	  Sleep(w->mlink.moveDelay / fast);
	}
	pos = w->mlink.spacePosition;
	for (aTile = 0; aTile < numTiles; aTile++) {
	    posNext = TileNFromSpace(w, aTile + 1, dir);
	    w->mlink.tileOfPosition[pos] =
	      w->mlink.tileOfPosition[posNext];
	    pos = posNext;
	}
	w->mlink.spacePosition = fillBeginPos;
	w->mlink.tileOfPosition[fillBeginPos] = 0;
}

static void
DiscreteMoves(MlinkWidget w, int from, int dir)
{
	mlinkCallbackStruct cb;

	MoveTiles(w, from, NORMAL);
	cb.reason = MLINK_MOVED;
	XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
	PutMove(dir, 0, 0);
}

static      Boolean
PositionTile(MlinkWidget w, int x, int *r)
{
	*r = (x - w->mlink.delta.x / 2 - w->mlink.puzzleOffset.x) / w->mlink.offset.x;
	if (*r < 0 || *r >= w->mlink.tiles)
		return False;
	return True;
}

static int
Column(MlinkWidget w, int pos)
{
	return (pos % w->mlink.tiles);
}

static int
Row(MlinkWidget w, int pos)
{
	return (pos / w->mlink.tiles);
}

static int
int2String(char *buf, int number, int base, Boolean capital)
{
	int         digit, mult = base, last, position;
	int         a, i, j, s, r;

	if (capital) {
		a = 'A', i = 'I', j = 'J', s = 'S', r = 'R';
	} else {
		a = 'a', i = 'i', j = 'j', s = 's', r = 'r';
	}
	if (number < 0) {
		char *buff;

		intCat(&buff, "int2String: 0 > ", number);
		XtWarning(buff);
		free(buff);
		return 0;
	}
	last = 1;
	while (number >= mult) {
		last++;
		mult *= base;
	}
	for (position = 0; position < last; position++) {
		mult /= base;
		digit = number / mult;
		number -= digit * mult;
		buf[position] = digit + '0';
		if (buf[position] > '9') {	/* ASCII */
			buf[position] += (a - '9' - 1);
		} else if (buf[position] < '0') {	/* EBCDIC */
			buf[position] += (a - '9' - 1);
			if (buf[position] > i)
				buf[position] += (j - i - 1);
			if (buf[position] > r)
				buf[position] += (s - r - 1);
		}
	}
	buf[last] = '\0';
	return last;
}

Boolean
CheckSolved(MlinkWidget w)
{
	int         i, j, firstTile, lastTile, midTile;

	if (w->mlink.tiles < 2)
		return True;
	if (w->mlink.orient) {
		firstTile = w->mlink.tileOfPosition[0];
		if (firstTile != 0 && (firstTile - 1) % w->mlink.tiles == 0) {
			for (i = 1; i < w->mlink.tileFaces; i++) {
				midTile = w->mlink.tileOfPosition[i];
				if (midTile && (midTile - firstTile + w->mlink.tileFaces) %
				    w->mlink.tileFaces != i)
					return False;
			}
		} else
			return False;
	} else {
		for (i = 0; i < w->mlink.faces; i++) {
			firstTile = w->mlink.tileOfPosition[i * w->mlink.tiles];
			if (firstTile == 0)
				firstTile = w->mlink.tileOfPosition[i * w->mlink.tiles + 1];
			lastTile = w->mlink.tileOfPosition[(i + 1) * w->mlink.tiles - 1];
			if (lastTile == 0)
				lastTile = w->mlink.tileOfPosition[(i + 1) * w->mlink.tiles - 2];
			if (firstTile % w->mlink.tiles != 1 ||
			    (lastTile % w->mlink.tiles != 0 &&
			     lastTile != w->mlink.tileFaces - 1) ||
			    (lastTile - 1) / w->mlink.tiles !=
			    (firstTile - 1) / w->mlink.tiles)
				return False;
			for (j = 1; j < w->mlink.tiles - 1; j++) {
				midTile = w->mlink.tileOfPosition[i * w->mlink.tiles + j];
				if ((midTile - 1) / w->mlink.tiles !=
				    (firstTile - 1) / w->mlink.tiles || midTile == 0)
					return False;
			}
		}
	}
	return True;
}
